/**
\mainpage
@defgroup hitran2refr refractive index calculation
@since 2001-07-18
@since 2009-01-10 option -D to adapt to database file relocations
@since 2010-01-20 added tcl GUI
@author Richard J. Mathar

This small collection of C++ classses calculates refractive indices
of air in the VIS to IR wavelength ranges, as published in
\latexonly
\cite{MatharAO43,MatharArxiv0610,MatharJOptA9}.
\endlatexonly

\section Compilation
Compilation in the style of
@code
make
@endcode
is supported with a corresponding makefile. The default assumes the availability
of the \c GSL libraries in the search paths for headers
and libraries; the corresponding compiler options in \c Makefile may
need adjustement. It is possible to compile the sources without availability
of these two libraries, but I do not have the time to adjust the code
or configuration to such needs.

\section Databases
The databases with line strengths come in various forms, not included
in hitran2refr.tgz,
<a href="http://cfa-www.harvard.edu/HITRAN">HITRAN 2004</a>,
<a href="http://spec.jpl.nasa.gov/ftp/pub/catalog">JPL</a> and
<a href="http://http://cfa-www.harvard.edu/amdata/ampdata/">the CFA</a>.
The location of these files is specified
in \c hitran2refr.cxx in the CPP definitions \c HITRAN2000DIR , \c SPECJPLDIR,
\c CFAWWWDIR and \c JCP111DIR . The directory with these directories
is
@code
..
@endcode
by default and can be specified with the
@code
-D
@endcode
option of the main executable.

Implementation and documentation come without any warranty and
are subject to quick and major changes at any point in the future.

<a href="http://www.strw.leidenuniv.nl/~mathar">Richard J. Mathar home page www.strw.leidenuniv.nl/~mathar</a>.
**/
using namespace std ;

#include <complex>
#include <sstream>
#include <fstream>
#include <iostream>
#include <iomanip>
#include <cmath>
#include <cstring>
#include <cstdlib>
#include <climits>
#include <unistd.h>

#include <set>
#include <vector>
#ifdef __GNUC__
#include <cstdarg>
#endif

#include <typeinfo>

#include "prconst.h"
#include "molecule.h"

#include "tips_2003.h"
#include "Air.h"


// define this to add the magnetic susceptibility
#define USE_MAGNETISM

// define this to add the temperature dependance of the partition sums
// The value of the macro is the HITRAN Reference temperature (K) for the partition sums
#define USE_QOFT 296


/** convert micron  to Terahertz
*/
#define MICRON2THZ(m) (VACUUMC*1.e-6/(m))

/** convert wavenumber to micron
*/
#define WN2MICRON(w) (1.e4/(w))

/** convert micron to wavenumber
*/
#define MICRON2WN(m) (1.e4/(m))

/** convert Celsius to Kelvin
*/
#define CELS2KELVIN(c) ((c)+273.15)


/** Print a short usage hint of the main program on standard output.
*/
void usage( char *argv0)
{
	cout << "usage: " << argv0 << " [-w] [-k] [-s count] [-P Pascal] [-T celsius] [-h relhum] lowlim(mu) hilim(mu) file1 ..\n" ;
	cout << "\tto produce the dispersion\n" ;
	cout << "\t-w uses wavenumbers (1/cm) not wavelengths on output and as hi/lowlim\n" ;
	cout << "\t-p remix partial pressures to attain fixed total pressure\n" ;
	cout << "\t-s sampling count\n" ;
}


/** Invocation from the UNIX shell.
\section Usage
@param[in] argc argument count, including command line options.
@param[in] argv the vector of all command line arguments.

The general call consists of a number of options followed by
two floating point arguments which are a spectral region in
either wavelength or wavenumber units, and a list of file names
in any of the few databases which correspond to molecular species.
@code
hitran2refr [options] range_low range_hi molec1 [molec2 ...]
@endcode

\subsection Command line options
-w If the option '-w' is used, the command line numbers of the spectral
   interval mean wavenumbers (1/cm), not wavelengths (micron). Also, the output is
   in wavenumbers, not microns.

-h <relhum> If the option '-h' followed by a number between 0 and 99
   is provided, this means a relative humidity in percent. The default
   is dry air (zero percent)

-s <sampl> If the option '-s' followed by a positive number is provided,
   this number is the number of sampling points on the wavelength
   (or with the -w option wavenumber) axis. The default is 2000.

-p If this option '-p' is added, the mixing ratios of all gas components
   will be adjusted to ensure that the gas pressures adds to the full
   pressure, even though the mixing ratios in the internal table
   do not add exactly to 100 percent.

-P If this option '-P' followed by a number is added, the number indicates
   the full gas pressure in units of Pa. The default is 75058.

-C The option '-C' followed by a floating point number specifies
   a volume density of the carbon dioxide in units of ppmv. The default
   is 370.

-T The option '-T' followed by a floating point number specifies
   the air temperature in units of Celsius.

-V disable the use of virial expansions. Use only the ideal gas law.
   The default is to use virial expansions.

-D The option '-D' followed by a directory name is the name
   of the directory which contains the directories HITRAN, spec and cfa
   with the databases of line transitions. This prime use is if the current
   working directory is not the directory of the original executable, where
   the databases are at the original place.
   The default is ".." (which indicates the directories ../HITRAN, ../spec
   and so on contain this information).

@return 0 on success and 1 on failure.
Failure indicates either use of invalid command line options or
command line parameters that are out of reasonable bounds.

\subsection Examples

Compute real and imaginary part of n-1 for a relative humidity of 20\%, a temperature
of 23 C, a total pressure of 750 hPa, and 1000 points for wavelengths from 3 to 6 micron,
using the hitran water line list in \c 011_hit06.par, adding the transitions
of the additional line list in \c h2ofvals.txt, the HITRAN nitrogen line list in \c 22_hit04.par,
the nitrogen line list in \c n2fvals.txt etc. Filter lines with an IEEE NaN output and
create the ASCII file \c tst.dat with the table.
@code
hitran2refr -s 1000 -h 20. -T 23 -p -P 75000 3.0 6.0 011_hit06.par h2ofvals.txt 22_hit04.par
n2fvals.txt 031_hit04.par 02_hit04.par co2fvals.txt 06_hit04.par 05_hit04.par 07_hit04.par
o2fvals.txt arfvals.txt nefvals.txt |sed '/NaN/d' | sed '/nan/d' > tst.dat
@endcode
*/
int main(int argc, char *argv[])
{

	if( argc < 2)
	{
		usage(argv[0]) ;
		exit(EXIT_FAILURE) ;
	}



	/* switch to indicate that the spectral range of the
	* command line arguments and the output is in wavenumbers (1/cm), not wavelengths
	*/
	bool x_waven=false ;

	/* Use virial expansions in place of the ideal gas law where available.
	* @since 2008-03-19
	*/
	bool useV =true ;

	/* switch to indicate that the mixing ratios of the dry 
	* components are adjusted/rescaled such that the full gas pressure at the
	* end equals the default or command line argument.
	*/
	bool scale_P=false ;
	int sampl= 0 ;		// undefined

	/* default total pressure is approximately 750 hPa
	*/
	double Pa= 75058.;

	/* default temperature is 12 C
	*/
	double T = CELS2KELVIN(12.);

	/* default humidity is 0 percent
	*/
	double relhum = 0. ;

	/* default CO2 fraction is 370 ppmv
	*/
	double co2 = 370.e-6 ;

	/* default O3 fraction is 0.1 ppmv
	* @since 2009-02-19
	*/
	double o3 = 1.e-7 ;

	/* default O2 fraction is 20.983 percent
	*/
#ifdef NOCO2
	double o2 = 0.20983 ;
#else
	double o2 = 0.20946 ;
#endif
	/** default is that h2o is set by relative humidity,  not
	* mixing ratio.
	*/
	double h2o = -1. ;

	/* range of wavelength in micron
	*/
	double mulimits[2] ;

	/* default directory for databases
	*/
	string dir("..") ;

	/* command line option character
	*/
	int c;
	while( (c=getopt(argc,argv,"ilwkbpVs:P:T:h:f:C:O:W:A:D:z:")) != -1 )
	{
		switch(c)
		{
		case 'w':
			x_waven=true ;
			break ;
		case 'h':
			relhum=atof(optarg) ;
			break ;
		case 's':
			sampl=atoi(optarg) ;
			break ;
		case 'p':
			scale_P=true ;
			break ;
		case 'V':
			useV=false ;
			break ;
		case 'P':
			Pa = atof(optarg) ;
			break ;
		case 'C':
			co2= 1.e-6*atof(optarg) ;
			break ;
		case 'D':
			dir= optarg ;
			break ;
		case 'O':
			o2= 1.e-6*atof(optarg) ;
			break ;
		case 'z':
			o3= 1.e-6*atof(optarg) ;
			break ;
		case 'W':
			h2o = 1.e-6*atof(optarg) ;
			break ;
		case 'T':
			T = CELS2KELVIN(atof(optarg)) ;
			break ;
		case '?':
			usage(argv[0]) ;
			exit(EXIT_FAILURE) ;
			break ;
		}
	}

	/** If the number of sampling points on the wavelngth or wavenumber axis
	* had not been provided by a command-line option, we chose some default.
	*/
	if ( sampl == 0 )
		sampl = 2000 ;

	{

#ifndef MOMENTS
		/** The two command line arguments following the options are interpreted
		* as the limits of the wavenumber or wavelength range of interest.
		*/
		if ( optind+1 >= argc)
		{
			cerr << "Command line arguments of wavelength or wavenumber range missing." << endl ;
			usage(argv[0]) ;
			exit(EXIT_FAILURE) ;
		}
		mulimits[0] = std::atof(argv[optind++]) ;
		mulimits[1] = std::atof(argv[optind++]) ;


		Air * paran ;
		Air::scale_p = scale_P;

		cout << "# " << T << " K, " << Pa << " Pa, " << relhum << " %\n" ;

		/** All command line arguments that follow the two number that specify the
		* spectral region are interpreted as individual base file names of one of
		* the databases, and are used to initialize the Air model with temperature,
		* pressure, relative humidity, CO2 content.
		*/
		paran= new Air(T,Pa,relhum,co2,o2,h2o,o3,useV,dir,argc-optind,argv+optind) ;

#ifdef USE_QOFT
		paran->Qsum() ;
#endif

		if( x_waven)
		{
			/* calculate the interval widths of looping over the spectral range */
			const double dk = (mulimits[1]-mulimits[0])/(sampl-1) ;

			cout << "# Columns: wavelength (micron), wavenumber (1/cm), 1/wavelength (1/micron), Re(n-1), Im(n), n(group)-1" << endl ;


			/* Loop over the wavenumber interval in equidistant steps.
			*/
			for( int kindx=0 ; sampl >0 ; kindx ++, sampl-- )
			{
				/* Compute current wavenumber and call air::suscep() with
				* the corresponding wavelength
				*/
				const double k=mulimits[0]+dk*kindx ;
				const double mu=WN2MICRON(k) ;
				complex<double> epsilon = 1.+paran->suscep(mu) ;

				/* add the effect of magnetism to the dielectric function
				*/
				epsilon *= 1.+paran->magn(useV) ;
				/* compute refractive index as square root of dielectric function
				*/
				complex<double> refrac= sqrt(epsilon) ;

				/* Produce one ASCII output line for this particular wavenumber, showing
				* lambda [mu], wavenumber [1/cm], inverse lambda [1/mu], Re(n-1), Im(n), Re(eps)-1, Im (eps).
				*/
				cout << mu << " " << MICRON2WN(mu) << " " << 1./mu << setprecision(12) << " " << real(refrac)-1. 
					<< " " << imag(refrac) << " " 
#if 0
					<< real(epsilon)-1. << " " << imag(epsilon) 
#endif
					<< endl ;
			}
		}
		else
		{
			/* compute step widths in the the spectral region (wavelength stride) 
			*/
			const double dmu = (mulimits[1]-mulimits[0])/(sampl-1) ;

				cout << "# Columns: wavelength (micron), wavenumber (1/cm), Re(n-1), Im(n), n(group)-1" << endl ;

			for( double mu= mulimits[0] ; sampl > 0; mu += dmu, --sampl )
			{
				complex<double> epsilon = 1.+paran->suscep(mu) ;
#ifdef USE_MAGNETISM
				epsilon *= 1.+paran->magn(useV) ;
#endif
				complex<double> refrac= sqrt(epsilon) ;

					/* Produce one ASCII output line for this particular wavelength, showing
					* lambda [mu], wavenumber [1/cm], inverse lambda [1/mu], Re(n-1), Im(n), Re(eps)-1, Im (eps).
					*/
					cout << mu << " " << MICRON2WN(mu) << setprecision(12) << " " << real(refrac)-1. << " " 
						<< imag(refrac) << " " 
#if 0
						<< real(epsilon)-1. << " " << imag(epsilon) 
#endif
						<< endl ;
			}
		}
		delete paran ;
#else /* MOMENTS */
#endif /* MOMENTS */
	}
	return 0 ;
}
