#!/bin/sh
# \
  exec wish "$0"  ${1+"$@"} >/dev/null </dev/null

# A GUI interface to the hitran2refr executable.
# see www.tcl.tk/man/tcl8.5/

# Richard J. Mathar
# 2009-04-30

# wm geometry . 400x900

# 1 if there are dry air components, else 0
proc hasdry {msel} {
	set nowat $msel
	set wat [lsearch $msel 0]
	if { $wat >= 0 } {
		set nowat [lreplace $nowat $wat $wat]
	}
	if { [llength $nowat] > 0 } {
		return 1
	} else {
		return 0
	}
}

# assemble the command line and execute it
proc go {} {
	global .lMol .tPress .tTemp .thum .tExec .tDBase .tout .dbg .tSamp .tMu .tco2 .to3

	set cmd [.tExec get -- 1.0 1.end]

	set P [expr [.tPress get -- 1.0 1.end] * 100]
	if { $P < 0 } {
		error "Negative pressure $P"
	}

	set cmdar -P
	lappend cmdar $P

	lappend cmdar -T [.tTemp get -- 1.0 1.end]

	set h [.thum get -- 1.0 1.end]
	if { $h < 0 || $h > 100 } {
		error "Invalid humidity $h"
	}
	lappend cmdar -h $h


	set s [.tSamp get -- 1.0 1.end]
	if { $s < 2 } {
		error "Invalid sampling count $s"
	}
	lappend cmdar -s $s

	set C [.tco2 get -- 1.0 1.end]
	set z [.to3 get -- 1.0 1.end]
	if { $C < 0 || $C > 1000000 || $z < 0 || $z > 1000000 } {
		error "Invalid density $C or $z"
	}
	lappend cmdar -C $C
	lappend cmdar -z $z
	lappend cmdar -D [.tDBase get -- 1.0 1.end]

	.dbg delete 1.0 end
	set msel  [.lMol curselection]

	# remove water from line list if zero humidity (speedup)
	if { $h == 0 } {
		set wat [lsearch $msel 0]
		if { $wat >= 0 } {
			set msel [lreplace $msel $wat $wat ]
		}
	}

	# remove CO2 from line list if zero density (speedup)
	if { $C == 0 } {
		set wat [lsearch $msel 1]
		if { $wat >= 0 } {
			set msel [lreplace $msel $wat $wat ]
		}
	}

	# remove O3 from line list if zero density (speedup)
	if { $z == 0 } {
		set wat [lsearch $msel 2]
		if { $wat >= 0 } {
			set msel [lreplace $msel $wat $wat ]
		}
	}

	# -p if something besides water is selected
	if { [hasdry $msel] } {
		lappend cmdar -p
	}
	scan [.tMu get -- 1.0 1.end] "%e %e" mulow muhi
	lappend cmdar $mulow $muhi

        set hdir  [.tDBase get -- 1.0 1.end]
	append hdir "/HITRAN/HITRAN2004/By-Molecule"

	set hifi [glob -directory $hdir -nocomplain -types f ??_hit??.par]

	# syntax check
	if { [llength $hifi] == 0 } {
		.dbg insert end "No HITRAN files found in $hdir"
		return
	}

	# add HITRAN data base files
	foreach f $hifi {
		set fbase [file tail $f] ; # base name like 14_hit08.par
		scan $fbase "%2d" idx ; # index like 14
		incr idx -1 ; # subtract 1 because the .lMol is zero-based, wheras file names are 1-based
		if { [lsearch $msel $idx] >= 0 } {
			lappend cmdar $fbase
		}
	}

	# add UV data base files
        set hdir  [.tDBase get -- 1.0 1.end]
	append hdir "/cfa"
	foreach f $msel {
		set fbase [string tolower [.lMol get $f]] ; # base name like h2o
		append fbase fvals.txt
		set hifi [glob -directory $hdir -nocomplain -types f $fbase]
		if { [llength $hifi] > 0 } {
			lappend cmdar $fbase
		}
	}

	lappend cmdar > [.tout get -- 1.0 1.end]
	.dbg insert end $cmd " " $cmdar
	.dbg insert end "\n"
	.dbg insert end "..running..."
	update
	eval exec $cmd $cmdar
	.dbg insert end "finished"
}

button .go -command {go} -state normal  -text {Start computation}

text .dbg -state normal -width 80 -height 14

listbox .lMol -selectmode multiple -selectbackground green

# the molecules exactly ordered according to their HITRAN enumeration
.lMol insert end "H2O" "CO2"  "O3" "N2O" "CO" "CH4" "O2" "NO" "SO2" "NO2" \
"NH3" "HNO3" "OH" "HF" "HCl" "HBr" "HI" "ClO" "OCS" "H2CO" \
"HOCl" "N2" "HCN" "CH3Cl" "H2O2" "C2H2" "C2H6" "PH3" "COF2" "SF6" \
"H2S" "HCOOH" "HO2" "O" "ClONO2" "NO+" "HOBr" "C2H4" "CH3OH" "CH3Br" \
"CH3CN" "CF4" "" "" "" "" "" "" "" "" \
"" "" "" "" "" "" "" "" "" "" \
"" "" "" "" "" "" "" "" "" "" \
"" "" "" "" "" "" "" "" "" "" \
"" "" "" "" "" "" "" "" "" "" \
"" "" "" "" "" "" "" "" "" "Ne" \
"Kr" "Ar" "H2" "He"

set lMolcnt [.lMol size ]

label .tMu_lbl -text "Wavelength range (2 numbers, micron):"
text .tMu -state normal  -height 1 -width 35
.tMu insert end "8.4 14.0"

label .tSamp_lbl -text "Abscissa points sampled:"
text .tSamp -state normal  -height 1 -width 10
.tSamp insert end "2000"

label .tPress_lbl -text "Total pressure (hPa):"
text .tPress -state normal  -height 1 -width 7
.tPress insert end "744.0"

label .thum_lbl -text "Humidity (%):"
text .thum -state normal  -height 1 -width 10
.thum insert end "10.0"

label .tTemp_lbl -text "Temperature (Cels):"
text .tTemp -state normal  -height 1 -width 10
.tTemp insert end "16.0"

label .tco2_lbl -text "CO2 (ppmv):"
text .tco2 -state normal  -height 1 -width 10
.tco2 insert end "378.0"

label .to3_lbl -text "O3 (ppmv):"
text .to3 -state normal  -height 1 -width 10
.to3 insert end "0.1"

# initally everything selected
# .lMol selection set 0 $lMolcnt
# rather not because there are too many command line arguments then...
.lMol selection set 0 12
.lMol selection set 21 21
.lMol selection set 99 $lMolcnt

label .lMol_lbl -text "Air composition (scrollable, green = selected):"

label .tExec_lbl -text "Executable:"
text .tExec -state normal  -height 1 -width 15
.tExec insert end "hitran2refr"

label .tDBase_lbl -text "Database directory:"
text .tDBase -state normal  -height 1 -width 15
.tDBase insert end "../"

label .tout_lbl -text "Output file:"
text .tout -state normal  -height 1 -width 35
.tout insert end "refr_tst"

button .quit -command {wm withdraw .; exit 0} -state normal  -text {Quit} -width 13

pack .tMu_lbl .tMu
pack .tPress_lbl .tPress
pack .tTemp_lbl .tTemp
pack .thum_lbl .thum
pack .lMol_lbl .lMol
pack .tExec_lbl .tExec
pack .tDBase_lbl .tDBase
pack .tout_lbl .tout
pack .tSamp_lbl .tSamp
pack .tco2_lbl .tco2
pack .to3_lbl .to3
pack .go
pack .dbg
pack .quit
