using namespace std ;

#include <complex>
#include <fstream>
#include <iostream>
#include <iomanip>
#include <cmath>
#include <cstdlib>
#include <unistd.h>
#include <typeinfo>

#include "prconst.h"

#include "transition.h"
#include "tips_2003.h"

/* 1-base definition of the indices of the atoms and molecules not part of the HITRAN database,
* in particular the noble gases, the hydrogen molecule etc
*/
#define ATOMS_OFFSET 100


/** convert Megahertz to wavenumber
*/
#define MHZ2WN(h) (1.e4*(h)/VACUUMC)

/** convert eV to wavenumber
*/
#define EV2WN(e) ((e)/HC)

/** convert wavenumber to hartree
*/
#define WN2HARTREE(w) (HC*(w)/HARTREE)

/** convert wavenumber to micron
*/
#define WN2MICRON(w) (1.e4/(w))

/** convert micron to wavenumber
*/
#define MICRON2WN(m) (1.e4/(m))

/** convert wavenumber to angular frequency
*/
#define WN2OMEGA(w) (2.*M_PI*100.*VACUUMC*(w))

/** Convert spectroscopic wavenumber [1/cm] to frequency [Hz] */
#define WN2HZ(w) (100.*VACUUMC*(w))

/** convert micron to angular frequency
*/
#define MICRON2OMEGA(m) (2.*M_PI*1.e6*VACUUMC/(m))

/** convert differential in micron to differential of angular frequency.
* \f$ d \omega /d \lambda = 2 \pi c/\lambda^2 \f$,
* \f$ d \omega=d \lambda \omega/\lambda = d \lambda \omega^2/(2\pi c)\f$ .
*/
#define DMICRON2DOMEGA(m,o) ((m)*(o)*(o)/(2.*M_PI*VACUUMC))

/** Obtain a formatted value from a HITRAN line (input stream).
* Read a formatted value of the indicated type from the input stream, but
* only to a maximum number of bytes
* @param[in,out] is the input file stream. On exit this is re-positioned to a point
*   \c width further than the position it had before calling the function.
* @param[out] buf the position to write the values read into
* @param[in] width maximum number of bytes to consider for input. This is
*  the decisive difference to the standard way of reading streams which would
*  swallow everything up to the next blank. This here is a reader made to
*  deal with Fortran-formatted, not necessarily blank-delimited lines of input.
*/
template <class TYPE> istream & getval(istream & is, TYPE & buf, int width)
{
	char userline[GEISALL] ;
	double tmpd ;
#ifdef __GNUC__
	is.get(userline,width+1) ;	// extract width bytes maximum
		/* convert fortran D characters of the GEISA implementation
		* if needed
		*/
	// if ( type_info(buf) == type_info(tmpd) )
	if ( typeid(buf) == typeid(tmpd) )
	{
		for(int p =0 ; p < width; p++)
			if ( userline[p] == 'D')
				userline[p] = 'E' ;
	}
	istringstream userstr(userline) ;
	userstr >> buf ;
#else
	streampos where = is.tellg() ;	// remember where we are
	is.getline(userline,GEISALL) ;	// extract GEISALL bytes maximum
	userline[width] = '\0' ;	// truncate temporary copy at 'width'
	istrstream userstr(userline) ;
	userstr >> buf ;
	is.seekg(where) ;		// reposition to old status, add 'width'
	is.seekg(width,ios::cur) ;
#endif
	return is;
}

/** Convert JPL molecular tags to HITRAN molecular numbers.
@param[in] jpltag one of the 5-digit JPL molecular ID's
@param[out] molec the HITRAN molecular number
@param[out] isot the HITRAN isotope number
@author Richard J. Mathar
@warning Only a limited list of molecules is handled.
*/
void tag2molec(const int jpltag, int &molec, int &isot)
{
	switch( jpltag)
	{
	case 18003:
	case 18005:
		molec=1 ; isot=1 ; break ;
	case 20003:
		molec=1 ; isot=2 ; break ;
	case 19003:
		molec=1 ; isot=3 ; break ;
	case 46013:
		molec=2 ; isot=3 ; break ;
	case 45012:
		molec=2 ; isot=4 ; break ;
	case 48004:
	case 48005:
	case 48006:
		molec=3 ; isot=1 ; break ;
	case 28001:
		molec=5 ; isot=1 ; break ;
	case 29001:
		molec=5 ; isot=2 ; break ;
	case 30001:
		molec=5 ; isot=3 ; break ;
	case 29006:
		molec=5 ; isot=4 ; break ;
	case 32001:
	case 32002:
		molec=7 ; isot=1 ; break ;
	case 34001:
		molec=7 ; isot=2 ; break ;
	case 33002:
		molec=7 ; isot=3 ; break ;
	case 46006:
		molec=10 ; isot=1 ; break ;
	case 16001:
		molec=34 ; isot=1 ; break ;
	}
}


/** default ctor.
* @param[in] molno molecular number
*/
transition::transition(int molno) : molec(molno), lowsta(0.)
{
}


/** Ctor with a string of a HITRAN line
* @param[in] in the ASCII line of one of the databases (catalogs)
* @param[in] cat the type of catalog. That is, the format and units in which
*   the columns of the input are organized.
*/
#ifdef __GNUC__
transition::transition (istringstream & in, catalog &cat)
#else
transition::transition (istream & in, catalog &cat)
#endif
{
	/* dummy temporary variable to store a floating point number
	*/
	double tmpf ;

	/* dummy temporary variable to store an integer
	*/
	int tmpi ;

	/* dummy temporary variable to store a string.
	*/
	string tmpa ;

	switch ( cat)
	{
	case HITRAN :
		/** For HITRAN input lines, recognize the I2, I1, F12.6 etc fortran specs
		* with molecule/isotope, wavenumber, strength, etc.
		*/
		getval(in,molec,2) ;		// Fortran I2
		getval(in,isot,1) ;		// Fortran I1
		getval(in,waven,12) ;		// Fortran F12.6 cm-1
		getval(in,strength,10) ;	// Fortran E10.3 transition strength
		getval(in,dipole,10) ;		// Fortran E10.3 Einstein A
		getval(in,width,5) ;		// Fortran F5.4
		getval(in,width_s,5) ;		// Fortran F5.4
		getval(in,lowsta,10) ;		// Fortran F10.4 cm-1

		/* the input was in wavenumbers, and we also store the wavelength
		*/
		micron=WN2MICRON(waven) ;
		break ;
	case SPECJPL:
		/** For JPL input lines, we consider frequencies, intensities etc, and
		* set the unspecified widths to zero.
		*/
		getval(in,tmpf,13) ;		// Fortran F13.4, FREQ, Megahertz
		waven=MHZ2WN(tmpf) ;
		getval(in,tmpf,8) ;		// Fortran F8.4, ERR
		getval(in,tmpf,8) ;		// Fortran F8.4, LGINT
		strength=pow(10.,tmpf)/(1.e10*VACUUMC) ;	// convert line intensity from nm^2*MHz to cm/molecule
		getval(in,tmpi,2) ;		// Fortran I2, DR
		getval(in,tmpf,10) ;		// Fortran F10.4, ELO
		getval(in,tmpi,3) ;		// Fortran I3, GUP
		getval(in,tmpi,7) ;		// Fortran I7, TAG
		tag2molec(tmpi,molec,isot) ;
		dipole=0. ;		// width unspecified
		lowsta=width=width_s=0. ;		// width unspecified
		micron=WN2MICRON(waven) ;
		break ;
	case CFAWWW:
		getval(in,molec,3) ;		// O3=3, N2 = 22, Ne=ATOMS_OFFSET, Kr= +1, Ar=+2, H2=+3, He=+4
		if( molec == ATOMS_OFFSET || molec == ATOMS_OFFSET+1 || molec == ATOMS_OFFSET+2 || molec == 2 || molec == 3)
		{
			isot=1 ;
			getval(in,tmpf,8) ;		// eV
			waven=EV2WN(tmpf) ;
			micron=WN2MICRON(waven) ;
			getval(in,f,10) ;
			if( molec == 2)
				f *= 18./1.37566;	// enforce f-sum=22  for total or 8 for outer electrons or 18 w/o O Kshells
			strength=dipole=width=width_s=0. ;
		}
		else if( molec == 1 || molec == 7 || molec==22 || molec == 6 || molec== 4 || molec== 11 
			|| molec== ATOMS_OFFSET+3)// JCP 68 p 1426
		{
			isot=1 ;
			in >> tmpf >> f ;	// Hartree and pseudo-oscillator
			waven=EV2WN(HARTREE*tmpf) ;
			micron=WN2MICRON(waven) ;
			strength=dipole=width=width_s=0. ;
		}
		else
		{
			isot=1 ;
			getval(in,tmpf,4) ;		// Angstroem
			micron=tmpf/1.e4 ;
			waven=MICRON2WN(micron) ;
			getval(in,f,10) ;
			strength=dipole=width=width_s=0. ;
		}

		/* no information on the position of the lower level
		*/
		lowsta=0. ;
		break ;
	case JCP111:

		/* only water stored in this format for now
		*/
		molec=1;
		isot=1 ;
		/* get wavenumber, 1/cm 
		*/
		getval(in,tmpf,12) ;
		micron=1.e4/tmpf ;
		waven=MICRON2WN(micron) ;
		/* get the relative (!) line strength
		*/
		getval(in,f,7) ;
		/* get a line width in milli inverse centimeters
		*/
		getval(in,tmpf,7) ;
		width=width_s=1.e-3*tmpf ;
		strength=f ;
		lowsta=dipole=0. ;
		break ;
	}

	/* rewind input line
	*/
	in.seekg(0) ;

	omega0=WN2OMEGA(waven) ;

	switch( cat)
	{
	case HITRAN:
	case SPECJPL:
		/** The conversion of line strengths follows as outlined by Hilborn, Am. J. Phys 50 (1981) 982,
		if the database does not provide one.
		\latexonly
		\cite{HilbornAJP50}
		\endlatexonly
		*/
		f= 0.04*EMASS*VACUUMC*VACUUMC*EPSILONZERO/(ECHARGE*ECHARGE)*strength ;
		break ;
	}
}

#ifdef USE_QOFT
/** If enable by the preprocessor macro, we addapt the line strength to the temperature.
\latexonly
\cite[(9)]{SimeckovaJQSRT98},
\cite[(5)]{JolyJQSRT102}.
\endlatexonly
*/
void transition::Qsum(const double T)
{
	/* factor exp[-h*nu/kB(1/T-1/T0)]
	*/
	double qf= -PLANCK*WN2HZ(lowsta)/BOLTZMANNK*(1./T-1./USE_QOFT) ;
	int gi ;
	qf=exp(qf)*bd_tips_2003(molec,USE_QOFT,isot,&gi)/bd_tips_2003(molec,T,isot,&gi)  ;
	qf *= (1.-exp(-PLANCK*WN2HZ(waven)/BOLTZMANNK/T))/(1.-exp(-PLANCK*WN2HZ(waven)/BOLTZMANNK/USE_QOFT)) ;
	strength *= qf; 
	f *= qf; 
}
#endif

/** Lorentz line profile.
* Note that this is \f$f/(\omega_0^2-\omega^2-i\gamma\omega)\f$, where $\gamma$
* is the FWHM, and replaced here by a line form that has the same integral (integrated
* line strength) but retains the symmetry of turning to the complex conjugated if the
* \f$\omega\f$ switches to its negative value.
* @param[in] omega
* @return the complex value at the position of \c omega .
*/
complex<double> transition::Lorentz(const double omega) const
{
	const complex<double> ihalf(0.,0.5) ;
	return 0.5*f*(1./(omega+omega0-ihalf*WN2OMEGA(width))-1./(omega-omega0-ihalf*WN2OMEGA(width)))/omega0 ;
}

/** First derivative of Lorentz line profile with respect to omega.
* @param[in] omega
* @return the complex value at the position of \c omega .
*/
complex<double> transition::LorentzPrime(const double omega) const
{
	const complex<double> ihalf(0.,0.5) ;
	return 0.5*f*(-1./pow(omega+omega0-ihalf*WN2OMEGA(width),2)+1./pow(omega-omega0-ihalf*WN2OMEGA(width),2))/omega0 ;
}

