#include <cmath>
#include <stdarg.h>
#include <cstdlib>
#include <iostream>
#include "prconst.h"
#include "virial.h"

using namespace std ;

/** default ctor.
* The default creator inmplies virials are essentially absent, which
* just gives the ideal gas behavior.
*/
virial::virial() : idealgas(true), ref(None)
{
}

/** Ctor.
* This constructor encodes a HITRAN molecule 
* @param[in] species the HITRAN integer encoding a molecular species
*/
virial::virial(const int species, const authors src ) : idealgas(false), molec(species), ref(src)
{
}

/** Convert number density into partial pressure.
* That is, convert moles per cubic meter into the partial pressure according to ideal or
* non-ideal gas equation.
* @param[in] mol_m3
* @param[in] T
* @return partial pressure [Pa]
*/
double virial::n2p(double mol_m3, const double T) const
{
		/* virial coefficients [m^3/mol] and [m^6/mol^2]
		*/
		double virsAB[2] ;	// virial coefficients in m^3/mol and m^6/mol^2

		/* Get the virial coefficients
		*/
		BC(T,virsAB[0],virsAB[1]) ;
		/* expand the virial series */
		return n2p(mol_m3,T,virsAB) ;
}

/** Convert partial pressure to number density.
* Inverse functionality of virial::n2p().
* @param[in] p partial pressure [Pa] 
* @param[in] T temperature [K]
* @return molecular density [moles/m^3]
*/
double virial::p2n(const double p, const double T) const
{
		/* virial coefficients [m^3/mol] and [m^6/mol^2]
		*/
		double virls[2] ;

		/* Get the virial coefficients
		*/
		BC(T,virls[0],virls[1]) ;

		/* expand the series
		*/
		return p2n(p,T,virls) ;
}

/** saturation density.
* @param[in] T
* @return saturation density [kg/m^3]
* @warning this is only implemented for the Wagner-Pruss model
*/
double virial::saturvap(const double T) const
{
	if ( ref == WagPruss )
	{
		/* critical density, equation (6.2) [kg/m^3] 
		*/
		const double rhoc= 322. ;
		const double theta= 1.-T/647.096 ;	// Tc from 2.2a
		// equation 2.7
		// also eq (3) in Wagner-Pruss JPhysChemRefData 22 p 783
		const double lgrat = -2.03150240*pow(theta,2./6.)-2.68302940*pow(theta,4./6.)
			-5.38626492*pow(theta,8./6.)-17.2991605*pow(theta,18./6.)-44.7586581*pow(theta,37./6.)
			-63.9201063*pow(theta,71./6.) ;
		return exp(lgrat)*rhoc ;
	}
	else
	{
		cerr << __FILE__ << " " << __LINE__ << " not implemented\n" ;
		exit(EXIT_FAILURE) ;
	}
}

/** Return virial coefficients.
* @param[in] T temperature [K]
* @return the value of B [m^3/mol] and the value of C [m^6/mol^2].
*/
void virial::BC(const double T, double &B, double &C) const
{
	/* For the ideal gas, both coefficients are zero.
	*/
	if ( idealgas)
		B=C=0. ;
	else
	{
		/* Virials of nitrogen, given as number triples with temperature (Kelvin), virial B (cm^3/mol)
		* and virial C (cm^6/mol^2)
		\latexonly
		\cite[p 239]{Dymond}
		\endlatexonly
		* @sa <a href="http://properties.nist.gov/SemiProp/Gases/N2.html#5">NIST</a>
		*/
		const double N2[][3]= {{263.08,-12.95,1590.},{276.94,-9.4,1500.},{291.41,-6.2,1460.}} ;
		const double O2[][3]= {
				/* virials for oxygen
				@see <a href="http://properties.nist.gov/SemiProp/Gases/O2.html#3">NIST</a>
				*/
				{270.,-22.57,1049.1},{280.,-20.,1008.5},{290.,-17.64,972.44},{285.,-19.24,1186.}} ;
		const double Ar[][3]= {
				/* virials for Argon.
				* @see <a href="http://properties.nist.gov/SemiProp/Gases/ar.html">NIST</a>
				*/
				{270.,-22.08,1129.6},{280.,-19.63,1094.0},{290.,-17.37,1062.2}} ;

		/* virials for carbon dioxide.
		* @see <a href="http://properties.nist.gov/SemiProp/Gases/co2.html">NIST</a>
		*/
		const double CO2[][3] = {{260.,-168.22,5411.3},{270.,-154.33,5255.4},{280.,-142.04,5079.8}} ;
		
		const double (*virtab)[3] =0;
		/* number of interpolatory points in the coefficients tables
		*/
		int n;

		/* virials are zero for molecules not considered above
		*/
		B=C=0. ;
		switch(molec)
		{
		case 2:
			virtab = CO2 ; n= sizeof(CO2)/sizeof(double[3]) ;
			break ;
		case 7:
			virtab = O2 ; n= sizeof(O2)/sizeof(double[3]) ;
			break ;
		case 22:
			virtab = N2 ; n= sizeof(N2)/sizeof(double[3]) ;
			break ;
		case 42:
			virtab = Ar ; n= sizeof(Ar)/sizeof(double[3]) ;
			break ;
		default:
			break ;
		}
		if ( virtab)
			/* Lagrange interpolation to the desired temperature \c T, given
			* T as \c virtab[...][0], li[0] for B, li[1] for C.
			\latexonly
			\cite[25.2.2]{AS}
			\endlatexonly
			*/
			for(int i=0 ; i < n ; i++)
			{
				double li=1. ; // li in 25.2.2 in AS, li[0] for B, li[1] for C
				for(int j=0; j < n; j++)
					if( j != i)
						li *= (T- virtab[j][0])/(virtab[i][0]-virtab[j][0]) ;
				B += li*virtab[i][1] ;
				C += li*virtab[i][2] ;
			}
		else if ( molec == 1)
		{
			if( ref == None || ref == HillMcMillan)
			{
				// Ind EngChem Res 27 (5) (1988), Table I, a0 to a6 and b0 to b2
				const double a[]={0.,0.160946868e-2,-0.1270588577e-2,-0.1947567798e-3,0.,0.,0.29604951749e-4};
				const double b[]={0.960620298296e15,0.,-.327058191623e21};
				// eq (16)
				double f[7] ;
				double alpha=10000. , beta=1500. ;
				f[1] = 1./(1.+T/alpha) ;
				f[3] = beta/T ;
				f[2] = pow(1.-exp(-f[3]),2.5)*sqrt(T/beta)*exp(f[3]) ;
				f[4] = f[3]*f[3] ;
				f[6] = f[4]*f[4] ;
				B = a[1]*f[1]+a[2]*f[2]+a[3]*f[3]+a[6]*f[6] ;	// have m^3/kg, convert kg->mol, m^3 to cm^3
				B *= WATERMOLWEIGHT*1000. ;	// now cm^3 per mol
				double tau = 1./T ;
				C = b[0]*pow(tau,7.)+b[2]*pow(tau,9.) ;	// eq (18), now have (m^3/kg)^2
				C *= pow(WATERMOLWEIGHT*1000.,2.) ;	// now cm^6/mol^2
			}
			else if ( ref == HarveyLemmon )
			{
				// incorporate Harvey J Phys Chem Ref Data 33 (1) (2004) // Eq 6., no extrapolation of Tab1
				const double Tstar= T/100. ;
				const double a[]={0.34404,-0.75826,-24.219,-3978.2} ;
				const double b[]={-0.5,-0.8,-3.35,-8.3};
				// eq (16)
				B = 0. ;
				for(int i=0; i < 4; i++)
					B += a[i]*pow(Tstar,b[i]) ;
				B *= 1.e3 ;
			}
			else if ( ref == WagPruss )
			{
				// incorporate WAgner J Phys Chem Ref Data, not  23 (1994) but 31 (2002)
				// parts of the Wagner-Pruss tables
				const struct { int ci; int di; double ti; double ni ; } tab62a[] = {
					{ 0, 1, -0.5, 0.12533547935523e-1},
					{ 0, 1, 0.875, 0.78957634722828e1},
					{ 0, 1, 1., -0.87803203303561e1},
					{ 0, 2, 0.5, 0.31802509345418},
					{ 0, 2, 0.75, -0.26145533859358},
					{ 0, 3, 0.375, -0.78199751687981e-2},
					{ 0, 4, 1., 0.88089493102134e-2},
					{ 1, 1, 4., -0.66856572307965},
					{ 1, 1, 6., 0.20433810950965},
					{ 1, 1, 12., -0.66212605039687e-4},
					{ 1, 2, 1., -0.19232721156002},
					{ 1, 2, 5., -0.25709043003438},
					{ 1, 3, 4., 0.16074868486251},
					{ 1, 4, 2., -0.40092828925807e-1},
					{ 1, 4, 13., 0.39343422603254e-6},
					{ 1, 5, 9., -0.75941377088144e-5},
					{ 1, 7, 3., 0.56250979351888e-3},
					{ 1, 9, 4., 0.156086522571353e-4},
					{ 1, 10, 11., 0.11537996422951e-8},
					{ 1, 11, 4., 0.36582165144204e-6},
					{ 1, 13, 13., -0.13251180074668e-11},
					{ 1, 15, 1., -0.62639586912454e-9},
					{ 2, 1, 7., -0.10793600908932},
					{ 2, 2, 1., 0.17611491008752e-1},
					{ 2, 2, 9., 0.22132295167546},
					{ 2, 2, 10., -0.40247669763528},
					{ 2, 3, 10., 0.58083399985759},
					{ 2, 4, 3., 0.49969146990806e-2},
					{ 2, 4, 7., -0.31358700712549e-1},
					{ 2, 4, 10., -0.74315929710341},
					{ 2, 5, 10., 0.47807329915480},
					{ 2, 6, 6., 0.20527940895948e-1},
					{ 2, 6, 10., -0.13636435110343},
					{ 2, 7, 10., 0.14180634400617e-1},
					{ 2, 9, 1., 0.83326504880713e-2},
					{ 2, 9, 2., -0.29052336009585e-1},
					{ 2, 9, 3., 0.38615085574206e-1},
					{ 2, 9, 4., -0.20393486513704e-1},
					{ 2, 9, 8., -0.16554050063734e-2},
					{ 2, 10, 6., 0.19955571979541e-2},
					{ 2, 10, 9., 0.15870308324157e-3},
					{ 2, 12, 8., -0.16388568342530e-4},
					{ 3, 3, 16., 0.43613615723811e-1},
					{ 3, 4, 22., 0.34994005463765e-1},
					{ 3, 4, 23., -0.76788197844621e-1},
					{ 3, 5, 23., 0.22446277332006e-1},
					{ 4, 14, 10., -0.62689710414685e-4},
					{ 6, 3, 50., -0.55711118565645e-9},
					{ 6, 6, 44., -0.19905718354408},
					{ 6, 6, 46., 0.31777497330738},
					{ 6, 6, 50., -0.11841182425981} } ;
				const struct { int di; int ti; double ni; int alphi; int beti; double gami; int epsi ;} tab62b[] = { 
					{ 3, 0, -0.31306260323435e2, 20, 150, 1.21 ,1 },
					{ 3, 1, 0.31546140237781e2, 20, 150, 1.21 ,1 },
					{ 3, 4, -0.25213154341695e4, 20, 250, 1.25 ,1 } } ;
				const struct { double ai; double bi; double Bi; double ni; int Ci; int Di; double Ai; double beti ;} tab62c[] = { 
					{ 3.5, 0.85, 0.2, -0.14874640856724, 28, 700, 0.32 , 0.3 },
					{ 3.5, 0.95, 0.2, 0.31806110878444, 32, 800, 0.32 , 0.3 } } ;

				// need with table 6.5 the first derivative of Phi at delta->0
				double tau = 647.096/T ;	// eq 2.2a
				double phideltar =0 ;
				for(int i=1; i<= 7; i++)
					if ( tab62a[i-1].di == 1)
						phideltar += tab62a[i-1].ni*tab62a[i-1].di*pow(tau,tab62a[i-1].ti) ;
				for(int i=8; i<= 51; i++)
					if ( tab62a[i-1].di == 1)
						phideltar += tab62a[i-1].ni*pow(tau,tab62a[i-1].ti)*tab62a[i-1].di ;
				// no contribution from there, because all di==3
				// for(int i=52; i<= 54; i++) ...
				for(int i=55; i<= 56; i++)
				{
						// at delta->0, eq 6.6 ...
					double psi = exp(-tab62c[i-55].Ci-tab62c[i-55].Di*(tau-1.)*(tau-1.)) ;
					// double dPsidd = -2.*tab62c[i-55].Ci*(-1.)*psi ; // bottom p 433 not needed
					double thet = 1.-tau+ tab62c[i-55].Ai ;	// after 6.6 where basis dleta-1=1
					double Delta = thet*thet + tab62c[i-55].Bi ;
					phideltar += tab62c[i-55].ni*pow(Delta,tab62c[i-55].bi) ;
				}
					
				const double rhoc= 322. ;	// (6.2) kg/m^3
					// divid by rhoc to get m^3/kg and multiply with kg/mol
				B = phideltar/rhoc*1.e3*WATERMOLWEIGHT ;	// now m^3/mole

				// need with table 6.5 the first derivative of Phi at delta->0
				phideltar =0 ;
				for(int i=1; i<= 7; i++)
					if ( tab62a[i-1].di == 2)
						phideltar += tab62a[i-1].ni*2.*tab62a[i-1].di*pow(tau,tab62a[i-1].ti) ;
				for(int i=8; i<= 51; i++)
					if ( tab62a[i-1].di == 2)
						phideltar += tab62a[i-1].ni*pow(tau,tab62a[i-1].ti)*tab62a[i-1].di ;
				// no terms with di=0 or di+(1..2)=0 or di=0 or di-2=0
				// for(int i=52; i<= 54; i++) ...
				for(int i=55; i<= 56; i++)
				{
						// at delta->0, eq 6.6 ...
					double psi = exp(-tab62c[i-55].Ci-tab62c[i-55].Di*(tau-1.)*(tau-1.)) ;
					double thet = 1.-tau+ tab62c[i-55].Ai ;	// after 6.6 where basis dleta-1=1
						// partial derivative D/ delta (p 434)
					double dDdd = -2.*(tab62c[i-55].Ai*thet*2/tab62c[i-55].beti
							+2.*tab62c[i-55].Bi*tab62c[i-55].ai) ;
					double Delta = thet*thet + tab62c[i-55].Bi ;
						// partial derivative D^bi/delta (third line bottom p 433)
					dDdd *= tab62c[i-55].bi*pow(Delta,tab62c[i-55].bi-1.) ;
						// partial derivative Psi/delta, bottom p 433
					double dPsidd = -2.*tab62c[i-55].Ci*(-1.)*psi ;
						// third line for Psi^r_{delta delta} for delta->0
					phideltar += tab62c[i-55].ni*( pow(Delta,tab62c[i-55].bi)*2.*dPsidd+2.*dDdd*psi ) ;
				}
					
					// divide by rhoc^2 to get m^6/kg^2 and multiply with kg^2/mol^2
				C = phideltar*pow(1.e3*WATERMOLWEIGHT/rhoc,2.) ;	// now m^6/mole^2

				return ;
			}
		}
		/* the units are now cm^3/mol and cm^6/mol^2. Convert them
		* to the final units to return.
		*/
		B *= 1.e-6 ;
		C *= 1.e-12 ;
	}
}

/** Convert number density to partial pressure.
* @param[in] mol_m3 number density [mol/m^3]
* @param[in] T temperature [K]
* @param[in] virls virial coefficents
* @return pressure [Pa]
*/
double virial::n2p(double mol_m3, const double T, const double virls[2]) const
{
		/* the virial expansion is pv/(RT)=1+B/v+C/v^2...
		*/
		return mol_m3*RMOLAR*T*(1.+mol_m3*virls[0]+mol_m3*mol_m3*virls[1]) ;
}

/** Convert partial pressure to number density.
* Inverse functionality of virial::n2p().
* @param[in] p partial pressure [Pa]
* @param[in] T temperature [K]
* @return density [mol/m^3]
*/
double virial::p2n(const double p, const double T, const double virls[2]) const
{
		/* Search for the n in p=RTn(1+n*virls[0]+n^2*virls[1])
		* Start with the guess from the ideal gas equation without virials
		*/
		double n = p/(RMOLAR*T) ;
		/* We use 4 Newton steps to solve
		* RTn + RTn^2*virls[0]+RTn^3*virls[1]-p=0
		*/
		for(int i=0 ; i < 4; i++)
			n -= (RMOLAR*T*n*(1.+n*virls[0]+n*n*virls[1])-p)/(RMOLAR*T*(1.+2.*n*virls[0]+3.*n*n*virls[1])) ;
		return n ;
}
