#include <iostream>
#include <cmath>
#include <cstdlib>

#include "water.h"
#include "WagnerPruss.h"
#include "prconst.h"

/** conversion factor grains/g */
#define GRAINSPERGRAMM 0.06479891

/** conversion factor feet per meter */
#define FEETPERMETER 0.3048

/** convert Farenheit to Kelvin
*/
#define FAREN2KELVIN(f) (((f)-32.)*5./9.+273.15)

/** convert Kelvin to Farenheit
*/
#define KELVIN2FAREN(k) (9.*((k)-273.15)/5.+32.)

using namespace std ;

// this call includes creating the ideal gas equation if the 'enhanc' parameter is 
// provided as Virial here...
water::water(double kelvin, enum eos enhanc) : T(kelvin), equa_of_s(enhanc)
{
	if ( enhanc != Virial)
		vir = 0 ;
	else
		vir = new virial() ;
}

/** Ctor
* @param[in] kelvin temperature [K]
* @param[in] v a virial for the deviations from the ideal gas equation
*/
water::water(double kelvin, virial &v) : T(kelvin), equa_of_s(Virial)
{
	vir = new virial(v) ;
}

/** Copy ctor.
*/
water::water(const water &orig)
{
	T = orig.T ;
	equa_of_s = orig.equa_of_s ;
	if ( orig.vir)
		vir = new virial(*orig.vir) ;
	else
		vir= 0 ;
}

/** Dtor.
*/
water::~water()
{
	delete vir ;
}

/** Compute water density.
* @param[in] humidity relative humidity in percent, from 0 to 100
* @return moles per cubic meter
*/
double water::n(const double humidity) const
{
	/* partial pressure
	*/
	double pascal ;
	switch(equa_of_s)
	{
	case Fairchild :
		return nFairchild(humidity) ;
	case WPruss :
		return nWPruss(humidity) ;
	case Virial :
		/* saturvap() returns kg/m^3. Divide by kg/mole to return moles per m^3
		* and multiply by 0.01 to convert percent to relative value.
		*/
		return humidity*0.01*vir->saturvap(T)/(WATERMOLWEIGHT*0.001) ;
	}
}

/** Compute water mixing ratio.
* @param[in] humidity relative humidity in percent
* @param[in] pres total pressure [Pa]
* @return mixing ration between 0 (dry) and 1
*/
double water::mixrat(const double humidity, const double pres ) const
{
	const double numDwat = n(humidity) ;	// water number density

	double pascal ;		// partial pressure of water (Pa)
	if ( vir )
		pascal = vir->n2p(numDwat,T) ;
	else
		pascal = RMOLAR*numDwat*T ;	// ideal gas equation

	const double numDair = (pres-pascal)/(RMOLAR*T) ; // ideal gas equation
	return numDwat/(numDwat+numDair) ;
}


/** Wagner-Pruss molecular density.
* @param[in] humidity
* @return moles per cubic meter
*/
double water::nWPruss(const double humidity) const
{
	WagnerPruss wp(T) ;
	return wp.n(humidity) ;
}

/** Water densities at the saturation point according to Fairchield
*/
double water::nFairchild(const double humidity) const
{

	/* Faren[] and densit[] are a list of temperatures [Farenheit] and densities
	* [grains per cubic feet?]
	*/
	const double Faren[]= { 0, 5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80, 85,
				90, 95 } ;
	const double densit[]= { 0.479, 0.613, 0.780, 0.988, 1.244, 1.558, 1.942, 2.375, 2.863, 3.436, 4.108,
		4.891, 5.800, 6.882, 8.066, 9.460, 11.056, 12.878, 14.951, 17.305 } ;
	const int tabs=sizeof(Faren)/sizeof(double) ;
	if( T < FAREN2KELVIN(Faren[0]) || T > FAREN2KELVIN(Faren[tabs-1]) )
	{
		cerr << "water::nFairchild() temperature " << T << " K out of range\n" ;
		exit(EXIT_FAILURE) ;
	}
	const double faren=KELVIN2FAREN(T) ;

	/* we search the appropriate interval that corresponds to the given temperature
	*/
	int indx ;
	for(indx=0 ; indx< tabs-1 ; indx++)
		if( Faren[indx] <= faren && Faren[indx+1] >= faren)
			break ;
	/* On a semi-logaritmic plot, we interpolate linearly between the
	* adjacent temperatures for the actual density
	*/
	const double c2= log(densit[indx]/densit[indx+1]) / ( Faren[indx]-Faren[indx+1] ) ;
	const double c1= log(densit[indx]) - c2 * Faren[indx] ;
	/* Multiply with GRAINSPERGRAMM/FEETPERMEETER^3 to obtain grams per cubic meter, then divide by
	* WATERMOLWEIGHT to obtain mols per cubic meter for dew point density, then multiply with Avogadro's constant
	* to obtain molecules per cubic meter; Finally multiply with percentage of humidity.
	*/
	return humidity*0.01*GRAINSPERGRAMM*exp(c1+c2*faren)
		/(FEETPERMETER*FEETPERMETER*FEETPERMETER*WATERMOLWEIGHT) ;
}

#undef FAREN2KELVIN
#undef KELVIN2FAREN

#undef GRAINSPERGRAMM
#undef FEETPERMETER
